
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '37.0.0rc2',
    'branch_nick': 'master',
    'revision_id': '6fb393d63e8f50733e00ce0b0234380ac12cd607',
    'revno': 13608
}
